/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2000, 2001, 2002, 2003 RiskMap srl
 Copyright (C) 2003, 2004, 2005, 2006 StatPro Italia srl
 Copyright (C) 2011 Ferdinando Ametrano
 Copyright (c) 2012 Klaus Spanderen

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <http://quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
*/

/*! \file observable.cpp
*/
#include <ql/errors.hpp>
#include <ql/patterns/observable.hpp>

#include <boost/signals2/signal.hpp>

using QuantLib::Observer;

namespace {
    class scoped_lock {
      public:
        scoped_lock(boost::signals2::mutex& mutex)
            : mutex_(mutex) { mutex_.lock();   }               
        ~scoped_lock()  { mutex_.unlock(); }
      private:
        boost::signals2::mutex& mutex_;
    };

    typedef char* pointer;
    std::set<void*, std::greater<void*> > globalSet;
    boost::signals2::mutex mutex;
        
    void addObserverToSet(void* obs) {
        scoped_lock lock(mutex);
        globalSet.insert(obs);
    }
    
    bool isObserver(void* obs) {
        scoped_lock lock(mutex);
        return globalSet.count(obs) > 0;
    }

    void setObserverThisPointer(void* obs) {
        scoped_lock lock(mutex);

        const std::set<void*, std::greater<void*> >::iterator iter 
            = globalSet.lower_bound(obs);

        if (iter != globalSet.end()) {
            pointer p = reinterpret_cast<pointer>(*iter);
            
            const std::size_t extra=2*(sizeof(std::size_t)+sizeof(Observer**));
            const std::size_t n = *reinterpret_cast<std::size_t*>(p-extra);
     
            if (obs < p + n) {
                *reinterpret_cast<Observer**>(p - 2*sizeof(Observer**)) 
                    = reinterpret_cast<Observer*>(obs);
            }
        }
    }

    std::size_t deleteObserverFromSet(void* obs) {
        scoped_lock lock(mutex);
        return globalSet.erase(obs);
    }
}


namespace boost {
    void sp_array_constructor_hook(void * p) {}
    void sp_array_destructor_hook(void * p) {}
    void sp_scalar_constructor_hook(void * p) {}
    void sp_scalar_constructor_hook( void * px, std::size_t size, void * pn ){}

    void sp_scalar_destructor_hook(void * p) {
        if (isObserver(p)) {
            const std::size_t extra = 2*sizeof(Observer**);
            (*reinterpret_cast<Observer**>(reinterpret_cast<pointer>(p)-extra))
                ->deactivate();
        }        
    }
    void sp_scalar_destructor_hook( void * px, std::size_t size, void * pn ){
        sp_scalar_destructor_hook(px);
    }
}

namespace QuantLib {

    class Observer::Proxy {
      public :
        Proxy(Observer* observer);
        void update();
        void deactivate();

        boost::signals2::mutex& mutex();
      private:
        bool active_;
        boost::signals2::mutex mutex_;
        Observer* const observer_;            
    };

    class Observable::Proxy {
      public:
        typedef boost::signals2::signal<void ()> signal_type;
        signal_type signal_;
    };
 
    Observable::Observable() 
    : proxy_(new Proxy()) {
    }

    Observable::Observable(const Observable&) 
    : proxy_(new Proxy()) {         
        // the observer set is not copied; no observer asked to
        // register with this object
    }

    /*! \warning notification is sent before the copy constructor has
                 a chance of actually change the data
                 members. Therefore, observers whose update() method
                 tries to use their observables will not see the
                 updated values. It is suggested that the update()
                 method just raise a flag in order to trigger
                 a later recalculation.
    */
    Observable& Observable::operator=(const Observable& o) {
        // as above, the observer set is not copied. Moreover,
        // observers of this object must be notified of the change
        if (&o != this)
            notifyObservers();
        return *this;
    }

    void Observable::registerObserver(
        const boost::shared_ptr<Observer::Proxy>& observerProxy) {

        proxy_->signal_.connect(
            Proxy::signal_type::slot_type(
                &Observer::Proxy::update, observerProxy.get())
            .track(observerProxy));
    }

    void Observable::unregisterObserver(
        const boost::shared_ptr<Observer::Proxy>& observerProxy) {

        proxy_->signal_.disconnect(
            boost::bind(&Observer::Proxy::update, observerProxy.get()));
    }

    void Observable::notifyObservers() {
        proxy_->signal_();
    }

    Observer::Proxy::Proxy(Observer* observer)
    : active_(true),
      observer_(observer) {
    }

    void Observer::Proxy::update() {
        scoped_lock lock(mutex_);
        if (active_) {
            observer_->update();
        }
    }

    void Observer::Proxy::deactivate() {
        scoped_lock lock(mutex_);
        active_=false;
    }

    boost::signals2::mutex& Observer::Proxy::mutex() {
        return mutex_;
    }

    Observer::Observer() 
      : proxy_(new Proxy(this)) {
        setObserverThisPointer(this);
    }

    Observer::~Observer() {
    }

    Observer::Observer(const Observer& o)
    : proxy_(new Proxy(this)), 
      observables_(o.observables_) {

        setObserverThisPointer(this);

        for (iterator i=observables_.begin(); i!=observables_.end(); ++i)
            (*i)->registerObserver(proxy_);
    }

    namespace {
        boost::signals2::mutex gobalEqualsLock;
    }
    Observer& Observer::operator=(const Observer& o) {
        proxy_.reset(new Proxy(this));

        scoped_lock g(gobalEqualsLock);
        scoped_lock l1(proxy_->mutex());
        {
            scoped_lock l2(o.proxy_->mutex());
            observables_ = o.observables_;
        }
        for (iterator i=observables_.begin(); i!=observables_.end(); ++i)
            (*i)->registerObserver(proxy_);
        return *this;
    }

    std::pair<std::set<boost::shared_ptr<Observable> >::iterator, bool>
    Observer::registerWith(const boost::shared_ptr<Observable>& h) {
        if (h) {
            h->registerObserver(proxy_);
            scoped_lock l(proxy_->mutex());
            return observables_.insert(h);
        }
        return std::make_pair(observables_.end(), false);
    }

    Size Observer::unregisterWith(const boost::shared_ptr<Observable>& h) {
        if (h)
            h->unregisterObserver(proxy_);
        scoped_lock l(proxy_->mutex());
        return observables_.erase(h);
    }

    void Observer::deactivate() {
        proxy_->deactivate();
    }

    void* Observer::operator new(std::size_t n) throw (std::bad_alloc) {
        const std::size_t extra = 2*(sizeof(std::size_t) + sizeof(Observer**));
        pointer const real_p = reinterpret_cast<pointer>(malloc(n+extra));
        if (real_p == 0) {
            throw std::bad_alloc();
        }
        pointer const p = real_p + extra;

        addObserverToSet(p);
        *(reinterpret_cast<std::size_t*>(real_p)) = n;

        return p;
    }

    void Observer::operator delete(void* p) {
        const std::size_t extra = 2*(sizeof(std::size_t) + sizeof(Observer**));
        pointer const real_p = reinterpret_cast<pointer>(p) - extra;
        
        deleteObserverFromSet(p);
        free(real_p);
    }
}
