/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2000, 2001, 2002, 2003 RiskMap srl
 Copyright (C) 2003, 2004, 2005, 2006 StatPro Italia srl
 Copyright (C) 2011 Ferdinando Ametrano
 Copyright (c) 2012 Klaus Spanderen

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <http://quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
*/

/*! \file observable.hpp
    \brief observer/observable pattern
*/

#ifndef quantlib_observable_hpp
#define quantlib_observable_hpp

#include <ql/types.hpp>
#include <ql/errors.hpp>

#include <boost/shared_ptr.hpp>
namespace boost {
    void sp_scalar_destructor_hook(void*);
}
#include <set>

namespace QuantLib {

    class Observable;

    //! Object that gets notified when a given observable changes
    /*! \ingroup patterns */
    class Observer {
        friend class Observable;
        friend void boost::sp_scalar_destructor_hook(void*);
      public:
        // constructors, assignment, destructor
        Observer();
        Observer(const Observer&);
        Observer& operator=(const Observer&);
        virtual ~Observer();

        // observer interface
        std::pair<std::set<boost::shared_ptr<Observable> >::iterator, bool>
                           registerWith(const boost::shared_ptr<Observable>&);
        Size unregisterWith(const boost::shared_ptr<Observable>&);
        /*! This method must be implemented in derived classes. An
            instance of %Observer does not call this method directly:
            instead, it will be called by the observables the instance
            registered with when they need to notify any changes.
        */
        virtual void update() = 0;

        void* operator new (std::size_t n) throw (std::bad_alloc);
        void operator delete (void* p);

      private:
        void deactivate();

        class Proxy; //pImpl idiom
        boost::shared_ptr<Proxy> proxy_;

        std::set<boost::shared_ptr<Observable> > observables_;
        typedef std::set<boost::shared_ptr<Observable> >::iterator iterator;
    };

    //! Object that notifies its changes to a set of observers
    /*! \ingroup patterns */
    class Observable {
        friend class Observer;
      public:
        // constructors, assignment, destructor
        Observable();
        Observable(const Observable&);
        Observable& operator=(const Observable&);
        virtual ~Observable() {}
        /*! This method should be called at the end of non-const methods
            or when the programmer desires to notify any changes.
        */
        void notifyObservers();

      private:
        void registerObserver(const boost::shared_ptr<Observer::Proxy>&);
        void unregisterObserver(const boost::shared_ptr<Observer::Proxy>&);

        class Proxy; //pImpl idiom
        boost::shared_ptr<Proxy> proxy_;
    };
}

#endif
